// =============================================================================
// CD-HIT
// http://bioinformatics.burnham-inst.org/cd-hi
// 
// program written by 
//                                      Weizhong Li
//                                      UCSD, San Diego Supercomputer Center
//                                      La Jolla, CA, 92093
//                                      Email liwz@sdsc.edu
//                 at
//                                      Adam Godzik's lab
//                                      The Burnham Institute
//                                      La Jolla, CA, 92037
//                                      Email adam@burnham-inst.org
// =============================================================================

#include "cd-hi.h"
#include "cd-hi-init.h"

////////////////////////////////////  MAIN /////////////////////////////////////
int main(int argc, char **argv) {
  int i, j, k, i1, j1, k1, i0, j0, k0, sggi, sgj;
  int si, sj, sk;
  char db_in[MAX_FILE_NAME];
  char db_out[MAX_FILE_NAME];
  char db_clstr[MAX_FILE_NAME];
  char db_clstr_old[MAX_FILE_NAME];

  // ***********************************    parse command line and open file
  if (argc < 5) print_usage(argv[0]);

  for (i=1; i<argc; i++) {
    if      (strcmp(argv[i], "-i") == 0)
      strncpy(db_in, argv[++i], MAX_FILE_NAME-1);
    else if (strcmp(argv[i], "-o") == 0)
      strncpy(db_out, argv[++i], MAX_FILE_NAME-1);
    else if (strcmp(argv[i], "-u") == 0) {
      strncpy(db_clstr_old, argv[++i], MAX_FILE_NAME-1);
      old_clstr_file = 1;
    }
    else if (strcmp(argv[i], "-M") == 0) 
      mem_limit = 1000000 * atoi(argv[++i]);
    else if (strcmp(argv[i], "-l") == 0) 
      length_of_throw = atoi(argv[++i]);
    else if (strcmp(argv[i], "-c") == 0) {
      NR_clstr = atof(argv[++i]);
      if ((NR_clstr > 1.0) || (NR_clstr < 0.4)) bomb_error("invalid clstr");
      NR_clstr100 = (int) (NR_clstr * 100 );
    }
    else if (strcmp(argv[i], "-b") == 0) {
      BAND_width = atoi(argv[++i]);
      if (BAND_width < 0 ) bomb_error("invalid band width");
    }
    else if (strcmp(argv[i], "-n") == 0) {
      NAA = atoi(argv[++i]);
      if ( NAA < 2 || NAA > 5 ) bomb_error("invalid word length");
    }
    else if (strcmp(argv[i], "-d") == 0) {
      des_len = atoi(argv[++i]);
      if ( des_len < 15 ) 
        bomb_error("too short description, not enough to identify sequences");
    }
    else if (strcmp(argv[i], "-t") == 0) {
      tolerance = atoi(argv[++i]);
      if ( tolerance < 0 || tolerance > 5 ) bomb_error("invalid tolerance");
    }
    else 
      print_usage(argv[0]);
  }
  db_clstr[0]=0; strcat(db_clstr,db_out); strcat(db_clstr,".clstr");

  if      ( NAA == 2 ) { NAAN = NAA2; }
  else if ( NAA == 3 ) { NAAN = NAA3; }
  else if ( NAA == 4 ) { NAAN = NAA4; }
  else if ( NAA == 5 ) { NAAN = NAA5; }
  else bomb_error("invalid -n parameter!");

  word_table.init(NAA, NAAN);

  if ( tolerance ) {
    int clstr_idx = (int) (NR_clstr * 100) - naa_stat_start_percent;
    int tcutoff = naa_stat[tolerance-1][clstr_idx][5-NAA];

    if (tcutoff < 5 ) 
      bomb_error("Too short word length, increase it or the tolerance");
    for ( i=5; i>NAA; i--) {
      if ( naa_stat[tolerance-1][clstr_idx][5-i] > 10 ) {
        cout << "Your word length is " << NAA << ", using "
             << i << " may be faster!" <<endl;
        break;
      }
    }
  }
  else {
    if      ( NR_clstr > 0.85 && NAA < 5) 
      cout << "Your word length is " << NAA
           << ", using 5 may be faster!" <<endl;
    else if ( NR_clstr > 0.80 && NAA < 4 )
      cout << "Your word length is " << NAA 
           << ", using 4 may be faster!" <<endl;
    else if ( NR_clstr > 0.75 && NAA < 3 )
      cout << "Your word length is " << NAA 
           << ", using 3 may be faster!" <<endl;
  }

  if ( length_of_throw <= NAA ) bomb_error("Too short -l, redefine it");

  ifstream in1(db_in);
  if ( ! in1 ) { cout << "Can not open file" << db_in << endl; exit(1); }
  ofstream out1(db_out);
  if ( ! out1) { cout << "Can not open file" << db_out << endl; exit(1); }
  ofstream out2(db_clstr);
  if ( ! out2) { cout << "Can not open file" << db_clstr << endl; exit(1); }

  DB_no = db_seq_no_test(in1); in1.open(db_in);
  if ((NR_len      = new int   [DB_no]) == NULL) bomb_error("Memory");
  if ((NR_idx      = new int   [DB_no]) == NULL) bomb_error("Memory");
  if ((NR90_idx    = new int   [DB_no]) == NULL) bomb_error("Memory");
  if ((NR_clstr_no = new int   [DB_no]) == NULL) bomb_error("Memory");
  if ((NR_iden     = new char  [DB_no]) == NULL) bomb_error("Memory");
  if ((NR_flag     = new char  [DB_no]) == NULL) bomb_error("Memory");
  if ((NR_seq      = new char *[DB_no]) == NULL) bomb_error("Memory");

  if ( old_clstr_file ) {
    ifstream in_clstr(db_clstr_old);
    if ( ! in_clstr) {
      cout << "Can not open file" << db_clstr_old << endl;
      exit(1);
    }

    //number of seq in old clstr file
    int clstr_seq_no = old_clstr_seq_no_test(in_clstr); 
    in_clstr.open(db_clstr_old);
  
    if ((NRo_idx       = new int [clstr_seq_no]) == NULL) bomb_error("Memory");
    if ((NRo_id1       = new int [clstr_seq_no]) == NULL) bomb_error("Memory");
    if ((NRo_id2       = new int [clstr_seq_no]) == NULL) bomb_error("Memory");
    if ((NRo_clstr_no  = new int [clstr_seq_no]) == NULL) bomb_error("Memory");
    if ((NRo_NR_idx    = new int [clstr_seq_no]) == NULL) bomb_error("Memory");
    if ((NRo_iden      = new char[clstr_seq_no]) == NULL) bomb_error("Memory");
    old_clstr_read_in(in_clstr, NRo_no, NRo90_no, NRo_idx, NRo_id1, NRo_id2, 
                      NRo_iden, NRo_clstr_no, NRo_NR_idx);
    in_clstr.close();
  }
  
  NRo_no > 0 ? 
    db_read_in2(in1, length_of_throw, NR_no, NR_seq, NR_len, 
                NRo_no, NRo_idx, NRo_id1, NRo_id2, NRo_NR_idx):
    db_read_in(in1, length_of_throw, NR_no, NR_seq, NR_len);
  in1.close(); 
  cout << "total seq: " << NR_no << endl;

  // ********************************************* init NR_flag
  for (i=0; i<NR_no; i++) NR_flag[i] = 0;
  if ( old_clstr_file ) {
    for (i=0; i<NRo_no; i++) {
      if ( (j = NRo_NR_idx[i]) == -1 ) continue ;
      if ( NRo_clstr_no[i] == i ) NR_flag[j] |= IS_OLD_REP;

      if ( (k = NRo_NR_idx[ NRo_clstr_no[i] ]) == -1 ) continue ;
      if ( NRo_clstr_no[i] != i ) NR_flag[j] |= IS_OLD_REDUNDANT;
      NR_iden[j] = NRo_iden[i];
      NR_clstr_no[j] = k; // note, later it need be changed to NR90_no
    }
    delete [] NRo_idx;
    delete [] NRo_id1;
    delete [] NRo_id2;
    delete [] NRo_iden;
    delete [] NRo_clstr_no;
    delete [] NRo_NR_idx;
  }

  sort_seqs_divide_segs(NR_no, NR_len, NR_idx, NR_seq, mem_limit, NAAN,
                        SEG_no, SEG_b, SEG_e, db_swap);

  // *********************************************                Main loop
  char *seqi;
  double aa1_cutoff = NR_clstr;
  double aa2_cutoff = 1 - (1-NR_clstr)*2;
  double aan_cutoff = 1 - (1-NR_clstr)*NAA;
  int len, hit_no, has_aa2, iden_no, aan_no, segb;
  int aan_list[MAX_SEQ];
  INTs aan_list_no[MAX_SEQ];
  INTs *look_and_count;
  if ((look_and_count= new INTs[NR_no]) == NULL) bomb_error("Memory");

  if ( tolerance ) {
    int clstr_idx = (int) (NR_clstr * 100) - naa_stat_start_percent;
    double d2  = ((double) (naa_stat[tolerance-1][clstr_idx][3]     )) / 100;
    double dn  = ((double) (naa_stat[tolerance-1][clstr_idx][5-NAA] )) / 100;
    aa2_cutoff = d2 > aa2_cutoff ? d2 : aa2_cutoff;
    aan_cutoff = dn > aan_cutoff ? dn : aan_cutoff;
  }

  NR90_no = 0;
  for (sggi=0; sggi<SEG_no; sggi++) {
    if (SEG_no >1)
      cout << "SEG " << sggi << " " << SEG_b[sggi] << " " << SEG_e[sggi] <<endl;

    for (sgj=sggi-1; sgj>=0; sgj--) {
      cout << "Reading swap" << endl;
      if ( sgj != sggi-1) word_table.read_tbl(db_swap[sgj]);    // reading old segment
      cout << "Comparing with SEG " << sgj << endl;
      for (i1=SEG_b[sggi]; i1<=SEG_e[sggi]; i1++) {
        i = NR_idx[i1];
        if (NR_flag[i] & IS_REDUNDANT  ) continue;

        if ( (NR_flag[i] & IS_OLD_REDUNDANT) &&
             (NR_flag[ NR_clstr_no[i] ] & IS_REP) ) {
          NR_clstr_no[i] = - (NR_clstr_no[ NR_clstr_no[i] ]) - 1;
          NR_flag[i] |= IS_REDUNDANT ;
          delete [] NR_seq[i];
          continue;
        }

        len = NR_len[i]; seqi = NR_seq[i];
        has_aa2 = 0;

        int flag = check_this(len, seqi, has_aa2,
               NAA, aan_no, aan_list, aan_list_no, look_and_count, 
               hit_no, SEG90_b[sgj], SEG90_e[sgj], iden_no,
               aa1_cutoff, aa2_cutoff, aan_cutoff,
               NR_flag[i], NR_flag);

        if ( flag == 1) {       // if similar to old one delete it
          delete [] NR_seq[i];
          NR_clstr_no[i] = -hit_no-1;  // (-hit_no-1) for non representatives
          NR_iden[i] = iden_no * 100 / len;
          NR_flag[i] |= IS_REDUNDANT ;
        }
      } //for (i1=SEG_b[sggi]; i1<=SEG_e[sggi]; i1++)
    } // for (sgj=0; sgj<sggi; sgj++)

    if (SEG_no >1) cout << "Refresh Memory" << endl;
    word_table.clean();

    if (SEG_no >1) cout << "Self comparing" << endl;
    segb = NR90_no;
    for (i1=SEG_b[sggi]; i1<=SEG_e[sggi]; i1++) {
      i = NR_idx[i1];
      
      if ( ! (NR_flag[i] & IS_REDUNDANT) ) {
        if ( (NR_flag[i] & IS_OLD_REDUNDANT) &&
             (NR_flag[ NR_clstr_no[i] ] & IS_REP) ) {
          NR_clstr_no[i] = - (NR_clstr_no[ NR_clstr_no[i] ]) - 1;
          NR_flag[i] |= IS_REDUNDANT ;
          delete [] NR_seq[i];
        }
        else {
          len = NR_len[i]; seqi = NR_seq[i];
          has_aa2 = 0;

          int flag = check_this(len, seqi, has_aa2,
                 NAA, aan_no, aan_list, aan_list_no, look_and_count, 
                 hit_no, segb, NR90_no-1, iden_no,
                 aa1_cutoff, aa2_cutoff, aan_cutoff,
                 NR_flag[i], NR_flag);

          if ( flag == 1) {       // if similar to old one delete it
            delete [] NR_seq[i];
            NR_clstr_no[i] = -hit_no-1;  // (-hit_no-1) for non representatives
            NR_iden[i] = iden_no * 100 / len;
            NR_flag[i] |= IS_REDUNDANT ;
          }
          else {                  // else add to NR90 db
            NR90_idx[NR90_no] = i;
            NR_clstr_no[i] = NR90_no; // positive value for representatives
            NR_iden[i] = 0;
            NR_flag[i] |= IS_REP;
            word_table.add_word_list(aan_no, aan_list, aan_list_no, NR90_no);
            NR90_no++;
          } // else
        } // else
      } // if ( ! (NR_flag[i] & IS_REDUNDANT) )

      if ( (i1+1) % 100 == 0 ) {
        cerr << ".";
        if ( (i1+1) % 1000 == 0 )
          cout << i1+1 << " finished\t" << NR90_no << " clusters" << endl;
      }
    } // for (i1=SEG_b[sggi]; i1<=SEG_e[sggi]; i1++) {

    SEG90_b[sggi] = segb;  SEG90_e[sggi] = NR90_no-1;
    if ( sggi < SEG_no-2 ) word_table.write_tbl( db_swap[sggi] ); // if not last segment
  } // for (sggi=0; sggi<SEG_no; sggi++) {


  for (i=0; i<NR90_no; i++)  delete [] NR_seq[ NR90_idx[i] ]; 

  cout << "writing new database" << endl;
  in1.open(db_in); 
  db_read_and_write(in1, out1, length_of_throw, des_len, NR_seq, NR_clstr_no);
  in1.close(); out1.close(); 

  cout << "writing clustering information" << endl;
  // write clstr information
  int *Clstr_no, *Clstr_size, *(*Clstr_list);
  if ((Clstr_no   = new int[NR90_no]) == NULL) bomb_error("Memory"); 
  if ((Clstr_size = new int[NR90_no]) == NULL) bomb_error("Memory"); 
  if ((Clstr_list = new int*[NR90_no]) == NULL) bomb_error("Memory");
  mem_size=5;
  for (i=0; i<NR90_no; i++) {
    Clstr_no[i]=0; Clstr_size[i]=mem_size;
    if((Clstr_list[i] = new int[mem_size]) == NULL) bomb_error("Memory");
  }

  for (i=0; i<NR_no; i++) {
    j1 = NR_clstr_no[i];
    if ( j1 < 0 ) j1 =-j1-1;
    if ( Clstr_no[j1] == Clstr_size[j1] ) {
      int *buff1;
      if ((buff1=new int[Clstr_size[j1]]) == NULL) bomb_error("Memory");
      for (k=0; k<Clstr_no[j1]; k++) buff1[k] = Clstr_list[j1][k];
      if ( Clstr_size[j1] >0 ) delete [] Clstr_list[j1];
      if ((Clstr_list[j1] = new int[mem_size+Clstr_size[j1]]) == NULL)
        bomb_error("Memory");
      for (k=0; k<Clstr_no[j1]; k++) Clstr_list[j1][k] = buff1[k];
      Clstr_size[j1]+=mem_size;
    }
    Clstr_list[j1][ Clstr_no[j1]++ ] = i;
  }

  for (i=0; i<NR90_no; i++) {
    out2 << ">Cluster " << i << endl;
    for (k=0; k<Clstr_no[i]; k++) {
      j = Clstr_list[i][k];
      out2 << k << "\t" << NR_len[j] << "aa, "<< NR_seq[j] << "...";
      if ( NR_iden[j]>0 ) out2 << " at " << int(NR_iden[j]) << "%" << endl;
      else                  out2 << " *" << endl;
    }
  }
  out2.close();
  cout << "program completed !" << endl << endl;

} // END int main

///////////////////////FUNCTION of common tools////////////////////////////


#include "cd-hit-g.c++"


/////////////////////////// END ALL ////////////////////////
